xquery version "3.1";
(:~
 : This module provides an interface to DokuWiki, transforms and stores exported
 : data including images from a configured wikispace. see config.xml.
 :
 : @author: Mathias Göbel
 : @version: 2.0
 : @see: https://fontane-nb.dariah.eu/doku.html
 :)

module namespace wiki="http://textgrid.de/ns/SADE/wiki";
import module namespace code="http://bdn-edition.de/ns/code-view";
import module namespace config="http://textgrid.de/ns/SADE/config" at "../config/config.xqm";
import module namespace functx="http://www.functx.com";
import module namespace rest="http://exquery.org/ns/restxq";

declare namespace http="http://expath.org/ns/http-client";
declare namespace output = "http://www.w3.org/2010/xslt-xquery-serialization";
declare namespace xhtml="http://www.w3.org/1999/xhtml";

(:~
 : returns the transformed wiki-text without a superior element (e.g. <div class="dokuwiki"/>)
 : @param $id A pagename in the wikispace according to config.xml
 : @param $node The current node form the html-template with context
 : @param $model The current execution model
 :)
declare function wiki:include($node as node(), $model as map(*), $id as xs:string)
as node()* {
  wiki:dokuwiki($node, $model, $id)/*
};

(:~
 : returns the transformed wiki-text
 : @param $id A pagename in the wikispace according to config.xml
 : @param $node The current node form the html-template with context
 : @param $model The current execution model
 :)
declare function wiki:dokuwiki($node as node(), $model as map(*), $id as xs:string)
 as node()* {
(: check for valid id and replace id not in whitelist :)
let $id:= if($id = doc('/db/sade-projects/textgrid/data/xml/data/2shfj.xml')//text()) then $id else 'gesamtdokumentation'
(: trigger image reload :)
let $imageLoad := if( request:get-parameter-names() = "reload" ) then wiki:dokuwikiImageReload($id) else false()

let $doc :=
        doc(
            '/sade-projects/textgrid/data/xml/doku/' || $id || '.rev' || wiki:revision-local($id) || '.xml' )/div/*

let $reload :=
    if( wiki:resource-is-locked($id) )
    then
        element xhtml:div {
            attribute class {"alert alert-info alert-dismissible"},
            <a href="#" class="close" data-dismiss="alert" aria-label="close">&#215;</a>,
            "Update in progress…"
        }
    else if(wiki:revision-remote($id) gt wiki:revision-local($id))
    then
        element xhtml:div {
                attribute class {"alert alert-info alert-dismissible"},
                <a href="#" class="close" data-dismiss="alert" aria-label="close">&#215;</a>,
                "Update in progress…",
                element xhtml:script {
                    "$.get('api/doku/update/" || $id || "').done(function() {location.reload();});"
                }
            }
    else if(not($doc))
    then
         error(QName("FONTANE", "DOKU06"), "Unable to get a local or remote document. Is the remote up and are there valid credentials available to the processor?" )
    else ()

return
  element xhtml:div {
    attribute class {
      "dokuwiki",
      $id
      },
    $reload,
    $doc
  }
};

(:~
 : Gets the currently stored revision number in the database for a given id
 : @param $id A pagename in the wikispace according to config.xml
 : @return the revision number stored in the database
 :)
declare function wiki:revision-local($id as xs:string)
as xs:int {
let $document-name as xs:string? :=
    xmldb:get-child-resources('/sade-projects/textgrid/data/xml/doku/')
        [substring-before(., '.rev') = $id]
        [not(ends-with(., ".lock.xml"))]

return
    if(not($document-name))
    then xs:int(-1)
    else
    $document-name
        => substring-after('.rev')
        => substring-before('.xml')
        => xs:int()
};

(:~
 : Gets the last revision number from the wikis feed and removes an old version
 : in the database. TODO: move remove operation somewhere else
 : @param $id The name of page in the wiki space
 : @param $lastRevinDb Revision number of the currently stored resource
 : @return the last revsion number from the remote wiki
 :)
declare function wiki:revision-remote($id as xs:string)
as xs:int {
let $wikiList :=
    try {
        ((xs:anyURI(config:get("dokuwiki.url")
        || '/feed.php?type=rss2&amp;num=500&amp;ns=fontane&amp;minor=1&amp;mode=list&amp;u='
        || config:get("dokuwiki.user")
        || '&amp;p='
        || config:get("dokuwiki.password"))
        => httpclient:get(false(), ()))/httpclient:body/node()
        )
    } catch * {
        error(QName("FONTANE", "DOKU05"), "Unable to connect to RSS feed. Are there valid credentials provided?" )
    }

let $lastRevinWiki :=
    ($wikiList//link
        [starts-with(., config:get("dokuwiki.url") || '/doku.php?id=fontane:'|| $id ||'&amp;rev=')]
        /substring-after(., '&amp;rev='))

return
    if( $lastRevinWiki )
    then $lastRevinWiki
    else xs:int(-1)
};

declare function wiki:lastModProcessor() {
    xmldb:last-modified("/db/apps/SADE/modules/wiki", "wiki.xqm")
};

declare function wiki:lock-resource($id as xs:string) {
    wiki:login(),
    xmldb:store('/sade-projects/textgrid/data/xml/doku/', $id || ".lock.xml", <lock/>)
};

declare function wiki:unlock-resource($id as xs:string) {
    wiki:login(),
    xmldb:remove('/sade-projects/textgrid/data/xml/doku/', $id || ".lock.xml")
};

declare %private function wiki:resource-is-locked($id as xs:string)
as xs:boolean {
    xmldb:get-child-resources('/sade-projects/textgrid/data/xml/doku/') = ($id || ".lock.xml")
};

(:~
 : Compares local and remote version number and performs an update if necessary.
 : Usually via AJAX call provided by the templating function: <code>wiki:dokuwiki()</code>
 : @param $id The name of page in the wiki space
 : @return <true/>
 :)
declare
  %rest:GET
  %rest:path("/api/doku/update/{$id}")
function wiki:reload($id as xs:string) {
let $remote := wiki:revision-remote($id)
let $local := wiki:revision-local($id)
return
    if( $remote le wiki:revision-local($id) )
    then
        error(QName("FONTANE", "DOKU01"), "Remote revision (" || $remote || ") is lower or equal local revision (" || $local || ")." )
    else (: proceed :)
    if ( wiki:resource-is-locked($id) )
    then
        error(QName("FONTANE", "DOKU02"), "Resource is locked." )
    else (: proceed :)
let $lock := wiki:lock-resource($id)
let $url as xs:anyURI :=
        (config:get("dokuwiki.url")
        || '/doku.php?id=fontane:' || $id
        || '&amp;rev=' || $remote
        ||'&amp;u='|| config:get("dokuwiki.user")
        ||'&amp;p='|| config:get("dokuwiki.password")
        ||'&amp;do=export_xhtml')
        => xs:anyURI(),
    $persist := false(),
    $request-headers:= (),
    $options:= <options><property name="http://cyberneko.org/html/properties/default-encoding" value="UTF-8"/></options>,
    $export :=
        try {
            httpclient:get($url, $persist, $request-headers, $options)//*[@class="dokuwiki export"]
        } catch * {
            error(QName("FONTANE", "DOKU04"), "unable to load resource." )
        },
    $result :=  if( $export//DIV )
                then wiki:HTML2html($export)
                else $export,
    $result := <div>{wiki:dokuwikiparser( $result/*[not(@id='dw__toc')] )}</div>,
    $login := wiki:login(),
    $remove := if($local = -1) then true() else
                xmldb:remove('/sade-projects/textgrid/data/xml/doku/', $id || '.rev' || $local || '.xml'),
    $store := xmldb:store( '/sade-projects/textgrid/data/xml/doku/', $id || '.rev' || $remote || '.xml' , $result ),
    $imageReload := wiki:dokuwikiImageReload($id)

let $unlock := wiki:unlock-resource($id)
return
    <true/>
};

(:~
 : Helper function to recursivly translate upper case element names to lower case
 :)
declare function wiki:HTML2html($node as node())
as node() {
  for $node in $node
  return
    typeswitch ( $node )
    case element() return
     element { lower-case($node/local-name()) } { $node/@*, wiki:HTML2html($node/node()) }
    default return $node
};

declare function wiki:dokuwikiparser($nodes as node()*){
    wiki:dokuwikiparser($nodes, false())
};

declare function wiki:dokuwikiparser($nodes as node()*, $char as xs:boolean){
let $char := false()
for $node in $nodes return
    typeswitch($node)
        case element(h1)
        return
            element xhtml:h2  {
                <div class="block-header">
                    <h2><span class="title">{wiki:dokuwikiparser($node/text(), $char)}</span><span class="decoration"></span><span class="decoration"></span><span class="decoration"></span></h2>
                    <div id="toggleToc"><i class="fa fa-caret-down" aria-hidden="true"></i></div>
                    <div id="dokuToc">
                        { wiki:dokuwikiTOC( $node/parent::div/div[@id="dw__toc"]/div/ul/li/ul) }
                    </div>
                </div>
            }
        case element(h2)
        return
            element xhtml:h3  {
                $node/@id,
                <span class="dokuLink"><a href="#{$node/@id}"> <i class="fa fa-link" aria-hidden="true"></i></a></span>,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(h3)
        return
            element xhtml:h4  {
                $node/@id,
                <span class="dokuLink"><a href="#{$node/@id}"> <i class="fa fa-link" aria-hidden="true"></i></a></span>,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(h4)
        return
            element xhtml:h5  {
                $node/@id,
                <span class="dokuLink"><a href="#{$node/@id}"> <i class="fa fa-link" aria-hidden="true"></i></a></span>,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(h5)
        return
            element xhtml:h6  {
                $node/@id,
                <span class="dokuLink"><a href="#{$node/@id}"> <i class="fa fa-link" aria-hidden="true"></i></a></span>,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(h6)
        return
            element xhtml:h7  {
                $node/@id,
                <span class="dokuLink"><a href="#{$node/@id}"> <i class="fa fa-link" aria-hidden="true"></i></a></span>,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case attribute(class) return ()
        case element(div)
        return
            element xhtml:div {(
                if($node/@class = "noteimportant")
                then
                    attribute class { 'note' }
                else (),
                wiki:dokuwikiparser($node/node(), $char)
            )}
        case element(br) return
            element xhtml:br {}
        case element(table)
        return
            element xhtml:table {
                attribute class {'table'},
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (thead)
        return
            element xhtml:thead {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (th)
        return
            element xhtml:th {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (tr)
        return
            element xhtml:tr {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (tbody)
        return
            element xhtml:tbody {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (td)
        return
            element xhtml:td {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(a)
        return
            if(exists($node/img)) then wiki:dokuwikiparser($node/node(), $char) else
            element xhtml:a {
                attribute href {
                    if($node/@class = "urlextern") then
                    (: wiki external links, should not link to test instance :)
                        attribute href {replace($node/@href, 'fontane-nb.dariah.eu/test/', 'fontane-nb.dariah.eu/')}
                    (: internal links :)
                    else if(starts-with($node/@href, '#'))
                        then string($node/@href)
                        else '?id='||substring-after($node/@href, 'fontane:')},
                $node/@rel,
                $node/@id,
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element(p)
        return
            element xhtml:p {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (ul)
        return
            element xhtml:ul {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (li)
        return
            element xhtml:li {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (pre)
        return
            $node
(:            let $code := try {:)
(:                string($node) => parse-xml():)
(:            } catch * {:)
(:                try {:)
(:                    ("<temporary-wrap>" || string($node) || "</temporary-wrap>") => parse-xml():)
(:                } catch * {<temporary-wrap>{string($node)}</temporary-wrap>}:)
(:            }:)
(:            return:)
(:                code:main($code) => wiki:remove-temporary-root() => functx:change-element-ns-deep("http://www.w3.org/1999/xhtml", ""):)

        case element (em)
        return
            element xhtml:em {
                if($node/@class = 'u') then attribute class {'underline'} else (),
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (strong)
        return
            element xhtml:strong {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (del)
        return
            element xhtml:del {
                wiki:dokuwikiparser($node/node(), $char)
            }
        case element (img)
        return
            (element xhtml:img {
                $node/@alt,
                $node/@width,
                $node/@title,
                attribute class {'imgLazy'},
                attribute data-original { 'public/doku/' || $node/substring-after(@data-original, 'fontane:') },
                attribute src {"public/img/loader.svg"}
            }, () )
        case text() return $node
        default return wiki:dokuwikiparser($node/node(), $char)
(:        case text() return $node:)
(:        default return wiki:dokuwikiparser($node/node()):)
};

declare function wiki:dokuwikiTOC($nodes) {
for $node in $nodes return
    typeswitch($node)
    case element(div) return
        wiki:dokuwikiTOC($node/node())
    case element() return
        element { $node/local-name() } { $node/@*[local-name() != 'class'], wiki:dokuwikiTOC($node/node()) }
    case text() return $node
    default return wiki:dokuwikiTOC($node/node())
};

declare function wiki:dokuwikiImageReload($id as xs:string) {
try {
let $path := system:get-exist-home() || '/tools/jetty/webapps/portal/public/doku/'
let $credentials := '&amp;u='|| config:get("dokuwiki.user") ||'&amp;p='|| config:get("dokuwiki.password")

let $url:= xs:anyURI(config:get("dokuwiki.url")||'/doku.php?id=fontane:'|| $id ||'&amp;do=export_xhtml' || $credentials),
    $persist := false(),
    $request-headers:= <http:header name="Accept-Charset" value="UTF-8"/>,
    $result:= httpclient:get($url, $persist, $request-headers)//div[@class="dokuwiki export"]/*[not(@id='dw__toc')]

let $images := distinct-values( $result//img/substring-after(@data-original, 'fontane:') )
return
    for $image in $images
    return
    let $id := $image
    let $req := <http:request href="{config:get("dokuwiki.url") ||"/lib/exe/fetch.php?media=fontane:"|| $id || $credentials }" method="get">
                <http:header name="Connection" value="close"/>
            </http:request>
    let $result :=  try {http:send-request($req)}
                    catch * {string($err:code)}
    let $mime := xs:string($result[1]//http:header[@name="content-type"]/@value)
    let $last-modified := xs:string($result[1]//http:header[@name="last-modified"]/@value)
    let $cache-control := xs:string($result[1]//http:header[@name="cache-control"]/@value)
    let $image := xs:base64Binary($result[2])
    let $login:= wiki:login()
    let $store := if( file:exists($path || $id)) then () else file:serialize-binary($image, $path || $id)
    return
        ()
} catch * {
    error(QName("FONTANE", "DOKU03"), "unable to (re-)load images." )

}
};

declare %private function wiki:login() {
if(false() = xmldb:login( '/sade-projects/textgrid/data/xml/doku/', config:get('sade.user') , config:get("sade.password")))
then (xmldb:login('/sade-projects/textgrid/data/xml/doku/', "admin", ""))
else true()
};

declare function wiki:remove-temporary-root($nodes) {
let $test := <xhtml:span class="hljs-tag">&lt;<xhtml:span class="hljs-name">temporary-wrap</xhtml:span>&gt;</xhtml:span>

for $node in $nodes
return
    typeswitch ( $node )
    case element( xhtml:span ) return
        if( deep-equal($node, $test) ) then ()
        else
        element {$node/node-name()} {
            $node/@*,
            wiki:remove-temporary-root($node/node())
        }
    case text() return
        normalize-space($node)
    default return
        element {$node/node-name()} {
            $node/@*,
            wiki:remove-temporary-root($node/node())
        }
};
